#from paper:: https://arxiv.org/pdf/2308.10882.pdf

import torch 

class RoPE:
    def __init__(
            self,
            d,
            k=None,
            a=None,
            b=None,
            rho=None
    ):
        self.d = d
        self.k = k
        self.a = a
        
        self.b = b
        self.rho = rho
    
    def power_scaling_basis(self):
        basis = torch.pow(10000 - 2 * torch.arange(self.d)[:, None] / self.d, 2 * 
                          torch.arange(self.d)[None, :] / self.d)
        modified_basis = (1 - 2 * torch.arange(self.d)[:, None] / 
                          self.d).pow(self.k) * basis
        return modified_basis
    
    def truncated_basis(self):
        basis = torch.pow(10000 - 2 * torch.arange(self.d)[:, None] / self.d, 2 * 
                          torch.arange(self.d)[None, :] / self.d)
        modified_basis = torch.where(basis >= self.b, basis, torch.where(basis > self.a, self.rho, 0))
        return modified_basis
    
#example usage
d = 10
k = 0.5
a = 0.1
b = 0.9
rho = 0.01

rope_power_scaling = RoPE(d=d, k=k)
modified_basis_power_scaling = rope_power_scaling.power_scaling_basis()
print(modified_basis_power_scaling)

rope_truncated = RoPE(d=d, a=a, b=b, rho=rho)
modified_basis_truncated = rope_truncated.truncated_basis()
print(modified_basis_truncated)


import numpy as np
import math
import torch

import einops
import torch.nn as nn
import torch.functional as F
from einops import rearrange

from typing import Callable, List, Optional, Tuple


####

def max_neg_values(tensor):
    return -torch.info(tensor.dtype).max

def l2norm(t, groups=1):
    t = rearrange(t, '... (g d) -> ... g d', g=groups)
    t = F.normalize(t, p=2, dim=-1)
    return rearrange(t, '... g d -> ... (g d)')

def pad_at_dim(t, pad, dim=-1, value=0.):
    dims_from_right = (- dim - 1) if dim < 0 else (t.ndim - dim - 1)
    zeros = ((0, 0) * dims_from_right)
    return F.pad(t, (*zeros, *pad), value=value)

def or_reduce(masks):
    head, *body = masks
    for rest in body:
        head = head | rest
    return head


class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn
    
    def forward(self, x, *args, **kwargs):
        return self.fn(x, *args, **kwargs) + x

class SinusoidalPosEmb(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim
    
    def forward(self, x):
        device = x.device
        half_dim = self.dim // 2
        
        emb = math.log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, device=device)* -emb)
        emb = x[:, None] * emb[None, :]

        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)
        return emb
    

def upsample(dim):
    return nn.ConvTranspose3d(dim, dim, (1, 4, 4), (1, 2, 2), (0, 1, 1))

def downsample(dim):
    return nn.Conv3d(dim, dim, (1, 4, 4), (1, 2, 2), (0, 1, 1))

class LayerNorm(nn.Module):
    def __init__(self, 
                 dim, 
                 eps=1e-5):
        super().__init__()
        self.eps = eps
        self.gamma = nn.Parameter(torch.ones(1, dim, 1, 1, 1))
    
    def forward(self, x):
        var = torch.vart(x, dim=1, unbiased=False, keepdim=True)
        mean = torch.mean(x, dim=1, keepdim=True)
        return (x - mean) / (var + self.eps).sqrt() * self.gamma
    

class PreNorm(nn.Module):
    def __init__(self,
                 dim,
                 fn):
        self.fn = fn
        self.norm = LayerNorm(dim)
    
    def forward(self, x, **kwargs):
        x = self.norm(x)
        return self.fn(x, **kwargs)
    

def cosine_beta_schedule(timesteps, s=0.008):
    steps = timesteps + 1
    x = torch.linspace(0, 
                       timesteps,
                       steps,
                       dtype=torch.float64)
    alphas_cumprod = torch.cos(((x / timesteps) + s) / (1 + s) * torch.pi * 0.5) ** 2
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
    return torch.clip(betas, 0, 0.9999)


class Normalize(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim
    
    def forward(self, x):
        return torch.nn.functional.normalize(x, dim=self.dim, p=2)
    
class LearnableLogitScaling(nn.Module):
    def __init__(
            self,
            logit_scale_init: float = 1 / 0.07,
            learnable: bool = True,
            max_logit_scale: float = 100,
    ) -> None:
        super().__init__()
        self.max_logit_scale = max_logit_scale
        self.logit_scale_init = logit_scale_init
        self.learnable = learnable

        log_logit_scale = torch.ones([]) * np.log(self.logit_scale_init)
        if learnable:
            self.log_logit_scale = nn.Parameter(log_logit_scale)
        else:
            self.register_bufffer("log_logit_scale", log_logit_scale)
    
    def forward(self, x):
        return torch.clip(self.logit_scale.exp(),
                          max=self.max_logit_scale) * x
    
    def extra_repr(self):
        st = f"logit_scale_init={self.logit_scale_init}, learnable={self.learnable}," \
            f"max_logit_scale={self.max_logit_scale}"
        return st
    


class EinOpsRearrange(nn.Module):
    def __init__(self, rearrange_expr: str,
                 **kwargs) -> None:
        super().__init__()
        self.rearrange_expr = rearrange_expr
        self.kwargs = kwargs
    
    def forward(self, x):
        assert isinstance(x, torch.Tensor)
        return einops.rearrange(x, self.rearrange_expr, **self.kwargs)
    

def cast_if_src_dtype(
        tensor: torch.Tensor, 
        src_dtype: torch.dtype, 
        tgt_dtype: torch.dtype
):
    updated = False
    if tensor.dtype == src_dtype:
        tensor = tensor.to(dtype=tgt_dtype)
        updated = True
    return tensor, updated


class SelectElements(nn.Module):
    def __init__(self,
                 index) -> None:
        super().__init__()
        self.index = index
    
    def forward(self, x):
        assert x.ndim >= 3
        return x[:, self.index, ...]
    

class SelectEOSAndProject(nn.Module):
    def __init__(self, proj: nn.Module) -> None:
        super().__init__()
        self.proj = proj
    
    def forward(self, x, seq_len):
        assert x.ndim == 3
        x = x[torch.arange(x.shape[0]), seq_len]
        x = self.proj(x)
        return x
    



##################
def get_sinusoid_encoding_table(n_position, d_hid):
    def get_position_angle_vec(position):
        return [
            position / np.power(10000, 2 * (hid_j // 2) / d_hid)
            for hid_j in range(d_hid)
        ]
    
    sinusoid_table = np.array(
        [get_position_angle_vec(pos_i) for pos_i in range(n_position)]
    )
    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2]) #dim 21
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])

    return torch.FloatTensor(sinusoid_table).unsqueeze(0)

def interpolate_pos_encoding_2d(target_spatial_size, pos_embed):
    N = pos_embed.shape[1]
    if N == target_spatial_size:
        return pos_embed
    dim = pos_embed.shape[-1]
    pos_embed, updated = cast_if_src_dtype(pos_embed, torch.bfloat16, torch.float32)
    pos_embed = nn.functional.interpolate(
        pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(
        0, 3, 1, 2
        ),
        scale_factor = math.sqrt(target_spatial_size / N),
        mode="bicubic",
    )
    if updated:
        pos_embed, _ = cast_if_src_dtype(pos_embed, torch.float32, torch.bfloat16)
    pos_embed = pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
    return pos_embed




import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms as T, utils
from PIL import Image

CHANNELS_TO_MODE = {
    1 : 'L',
    3 : 'RGB',
    4 : 'RGBA'
}


def seek_all_images(img, channels = 3):
    assert channels in CHANNELS_TO_MODE, f'channels {channels} invalid'
    mode = CHANNELS_TO_MODE[channels]

    i = 0
    while True:
        try:
            img.seek(i)
            yield img.convert(mode)
        except EOFError:
            break
        i += 1


#tensor of shape (channels, frames, height, width) -> GIF
def video_tensor_to_gift(tensor, path, duration=120, loop=0, optimize=True):
    images = map(T.ToPilImage(), tensor.unbind(dim=1))
    first_img, *rest_imgs = images
    first_img.save(path, 
                   save_all=True,
                   appeqnd_images=rest_imgs,
                   duration=duration,
                   loop=loop,
                   optimize=optimize
                   )
    return images


#gif -> (channels, frame, height, width) tensor
def gif_to_tensor(path,
                  channels=3,
                  transform=T.ToTensor()):
    img = Image.open(path)
    tensors = tuple(map(transform, seek_all_images(img, chanels=channels)))
    return torch.stack(tensors, dim=1)

def identity(t, *args, **kwargs):
    return t

def normalize_img(t):
    return t * 2 - 1

def unnormalize_img(t):
    return (t + 1) * 0.5

def cast_num_frames(t, *, frames):
    f = t.shape[1]

    if f == frames:
        return t
    
    if f > frames:
        return t[:, :frames]
    
    return F.pad(t, (0, 0, 0, 0, 0, frames - f))







import torch
import torch.nn as nn
from accelerate import Accelerator
from einops import rearrange

from zeta.nn.utils.helpers import exists



def print_num_params(model, accelerator: Accelerator):
    # n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    accelerator.print(f"Number of parameters in model: {n_params}")



class Block(nn.Module):
    def __init__(self,
                 dim,
                 dim_out,
                 groups=8):
        super().__init__()
        self.proj = nn.Conv3d(dim, dim_out, (1, 3, 3), padding=(0, 1, 1))
        self.norm = nn.GroupNorm(groups, dim_out)
        self.act = nn.SiLU()
    
    def forward(self, 
                x,
                scale_shift=None):
        x = self.proj(x)
        x = self.norm(x)

        if exists(scale_shift):
            scale, shift = scale_shift
            x = x * (scale + 1) + 1

        return self.act(x)


class ResnetBlock(nn.Module):
    def __init__(self,
                 dim,
                 dim_out,
                 *,
                 time_emb_dim=None,
                 groups=8):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.SiLU(),
            nn.Linear(time_emb_dim, dim_out * 2)
        ) if exists(time_emb_dim) else None
        
        self.block1 = Block(dim, dim_out, groups=groups)
        self.block2 = Block(dim_out, dim_out, groups=groups)
        self.res_conv = nn.Conv3d(dim, dim_out, 1) if dim != dim_out else nn.Identity()

    def forward(self,
                x,
                time_emb=None):
        scale_shift = None
        if exists(self.mlp):
            assert exists(time_emb), 'time_emb must be passed in'
            time_emb = self.mlp(time_emb)
            time_emb = rearrange(time_emb, 'b c -> b c 1 1 1')
            scale_shift = time_emb.chunk(2, dim=1)
        
        h = self.block1(x, scale_shift=scale_shift)
        h = self.block2(h)
        
        return h + self.res_conv(x)
    
def load_model(path):
    with open(path, 'rb') as f:
        return torch.load(f, map_location=torch.device('cpu'))
    



# Copyright (c) 2022 Agora
# Licensed under The MIT License [see LICENSE for details]

#architecture
from zeta.nn.architecture.transformer import (
    AttentionLayers,
    Decoder,
    Encoder,
    Transformer,
    ViTransformerWrapper,
)


#attentions
from zeta.nn.attention.multiquery_attention import MultiQueryAttention
from zeta.nn.attention.flash_attention import FlashAttention
from zeta.nn.attention.dilated_attention import DilatedAttention

#models
from zeta.models.gpt4 import GPT4, GPT4MultiModal
from zeta.models.andromeda import Andromeda


#######
from zeta.nn import *
from zeta.models import *
from zeta.training import *



#training

#tokenizers
from zeta.training.multi_modal_tokenizer import MultiModalTokenizer
from zeta.training.language_tokenizer import LanguageTokenizerGPTX

#train 
from zeta.training.train import Trainer, train


#loss
from zeta.training.loss.nebula import Nebula


# Copyright (c) 2022 Agora
# Licensed under The MIT License [see LICENSE for details]

######### Attention
from zeta.nn.attention.multiquery_attention import MultiQueryAttention
from zeta.nn import *
from zeta.nn.attention.cross_attention import CrossAttend
from zeta.nn.attention.multihead_attention import MultiheadAttention
from zeta.nn.attention.flash_attention2 import FlashAttentionTwo




#utils
from zeta.nn.utils.utils import *
from zeta.nn.utils.helpers import *
from zeta.nn.utils.inference_helpers import *
from zeta.nn.utils.model_utils import *
from zeta.nn.utils.tensor_helpers import *
from zeta.nn.utils.multi_modal_helpers import * 



from math import ceil

import torch
import torch.functional as F
from torch import nn


def top_p(logits, thres = 0.9):
    sorted_logits, sorted_indices = torch.sort(logits, descending=True)
    cum_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)

    sorted_indices_to_remove = cum_probs > (1 - thres)
    sorted_indices_to_remove[:, 1:] = sorted_indices_to_remove[:, :-1].clone()
    sorted_indices_to_remove[:, 0] = 0

    sorted_logits[sorted_indices_to_remove] = float("-inf")
    return sorted_logits.scatter(1, sorted_indices, sorted_logits)


def top_k(logits, thres=0.9):
    k = ceil((1 - thres) * logits.shape[-1])
    val, ind, = torch.topk(logits, k)
    probs = torch.full_like(logits, float("-inf"))
    probs.scatter_(1, ind, val)
    return probs

def top_a(
        logits,
        min_p_pow=2.0,
        min_p_ratio=0.02
):
    probs = F.softmax(logits, dim=-1)
    limit = torch.pow(torch.max(probs), min_p_pow) * min_p_ratio

    logits[probs < limit] = float("-inf")
    logits[probs >= limit] = 1
    return logits

def log(t, eps=1e-20):
    return torch.log(t.clamp(min=eps))

def gumbel_noise(t):
    noise = torch.zeros_like(t).uniform_(0, 1)
    return -log(-log(noise))

def gumnel_sample(t, temperature=1., dim=-1):
    return ((t / max(temperature, 1e-10)) + gumbel_noise(t)).argmax(dim=dim)


class ContrastiveTopK(nn.Module):
    def __init__(self, 
                 alpha,
                 k):
        super(ContrastiveTopK, self).__init__()
        self.alpha = alpha
        self.k = k
    
    def top_k(self, logits):
        k = ceil((1 - self.alpha) * logits.shape[-1])
        val, ind = torch.topk(logits, k)
        
        probs = torch.full_like(logits, float('-inf'))
        probs.scatter_(1, ind, val)

        return probs
    
    def forward(self,
                logits_exp,
                logits_ama):
        logits_exp_topk = self.top_k(logits_exp)
        logits_ama_topk = self.top_k(logits_ama)

        #probabilities
        p_exp = F.softmax(logits_exp_topk, dim=-1)
        p_ama = F.softmax(logits_ama_topk, dim=-1)

        #mask
        _, ind = torch.topk(p_exp, self.k)
        mask = torch.zeros_like(p_exp)
        mask.scatter_(1, ind, p_exp[ind] >= self.alpha * p_exp[ind[-1]])

        #scores
        scores = torch.where(mask.bool(), torch.log(p_exp / (p_ama + 1e-8)), 
                             torch.tensor(-float('inf')))
        
        return scores

#alpha = 0.5
#k = 10
# cdk = ContrastiveTopK(alpha, k)

#logits_exp = torch.randn(100, 50)
#logits_ama = torch.randn(100, 50)

#scores
#scores = cdk(logits_exp, logits_ama)
#return